/* text.c test program for game_lib.c
* 
* copyright 06/26/2001
* dave blood - geekd@yahoo.com
*
* 07/03/2001 - Mike Heckman
*  Added pause capability and compiler options to turn off sound 
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "gamelib.h"
#include "DT_drawtext.h"
#include "SDL_image.h"
 
#define XSIZE 640
#define YSIZE 480
#define TERRAIN_YSIZE (YSIZE / 2)
#define FPS (1000 / 35)

typedef struct {
  int r;
  int g;
  int b;
} RGBcolor;

typedef struct {
  int ships_remaining;
  unsigned int fuel;
  int score;
  int difficulty;
  float gravity;
  int big_font;
  int small_font;
} Game;


typedef struct {
  Sprite   sprite;
  int      landing_x[5];
  int      landing_y[5];
  int      landing_w[5];
  float    landing_speed[5];
  int      landing_score[5];
  RGBcolor landing_color[5];
  int      difficulty;
  int      fuel;
  int      num_landings;
} Level;


void remaining_ships( SDL_Surface *screen, 
		      Sprite miniship, 
		      Game game ) {
  int count;

  // Draw in the remaining ships
  for( count = 0; count < game.ships_remaining; count++ ) {
    miniship.x = (22 * count);
    miniship.y = 15;
    draw_sprite(screen, miniship );
  }

}



void DrawPixel(SDL_Surface *screen, Uint8 R, Uint8 G, Uint8 B, int x, int y)
{

  Uint32 color = SDL_MapRGB(screen->format, R, G, B); 
  Uint16 *bufp;
  bufp = (Uint16 *)screen->pixels + y*screen->pitch/2 + x;
  *bufp = color;

}

void draw_line(SDL_Surface *screen, Uint8 R, Uint8 G, Uint8 B, int x, int y) {
  
  for(; y < TERRAIN_YSIZE; y++){
    DrawPixel(screen, R, G, B, x, y);
  }

}

Level random_level(Game game) {

  SDL_Surface *terrain;
  Sprite sprite;
  Level level;
  int x = 0;
  int y = 0;
  int yd = 0;
  int miny = 0;
  int maxy = 0;
  int yfluct = 0;
  int size = 0;
  int max = 0;
  float speed;
  RGBcolor rgb;
  float distance;
  int xdiff;
  int ydiff;

  level.num_landings = 0;
  level.difficulty = game.difficulty;
  
  while (level.num_landings == 0){

    terrain = IMG_Load("images/blank_terrain.bmp");
    terrain = SDL_DisplayFormat(terrain); 

    SDL_LockSurface(terrain);

    srandom(time(NULL)); 

    while ( (y > 230) || (y < 10) ){
      y = random()%TERRAIN_YSIZE;
    }

    miny = ( (random()%(TERRAIN_YSIZE/2)) + 10);
    maxy = ( (random()%(TERRAIN_YSIZE/2) + TERRAIN_YSIZE/2) -10 );

    if (y > (TERRAIN_YSIZE/2)){
      yd = -1;
    }
    else{
      yd = 1;
    }


    /* start to draw */
  
    for (x = 0; x < XSIZE; x++){

      if ( (x % 3) == 0){
	yfluct = ( (random()%10) - 5);
      }

      if ( ( (y < maxy) && (y > miny)) || ( (y > maxy) && (yfluct < 1) ) 
	   || ( (y < maxy) && (yfluct > -1) ) ){
    
	y = y + yfluct;

      }

      /* draw landscape */
    
      draw_line(terrain, 80, 80, 80, x, y);

      /* check for change direction */
    
      if (yd == 1){

	if (y > maxy){

	  if (x < (XSIZE - 60) ){

	    /* landing pad */

	    size = (40 + (15 - ( game.difficulty / 2 ) ) );
	    max = x + size;

	    level.landing_x[level.num_landings] = x;
	    level.landing_y[level.num_landings] = (y + TERRAIN_YSIZE);;
	    level.landing_w[level.num_landings] = size;

            /* get distance from center-top */
            xdiff = ((XSIZE / 2) -  x );
            ydiff = ( y );

            if (xdiff == 0){
              distance = ydiff;
            }
            else if (ydiff == 0){
              distance = xdiff;
            }
            else {
              distance = sqrt( ( xdiff * xdiff  ) + (ydiff * ydiff ) );
	    }

	    level.landing_speed[level.num_landings] = 1.00;
	    level.landing_score[level.num_landings] = 
	      ( ( (16 - (size - 45) ) * 100) + 
		( (distance / 4) * ( (game.difficulty / 2) + 1) ) );

	    // Default landing pad color is blue
	    rgb.r = 0;
	    rgb.g = 0;
	    rgb.b = 255;
	    
	    if( level.num_landings ) {
	      speed = ( random() % 3 );

	      // Magenta is for landings at .90
	      if( speed > 0  ) {
		rgb.r = 255;
		rgb.g = 0;
		rgb.b = 255;
	      }
	      // Green for landings at .80
	      if( speed > 1 ) {
		rgb.r = 0;
		rgb.g = 255;
		rgb.b = 0;
	      }
	      
	      // Set the landing speed and score
	      level.landing_speed[level.num_landings] -= (speed * .1);
	      level.landing_score[level.num_landings] += (speed * 100);
	    }	      
	    level.landing_color[level.num_landings] = rgb;

	    // Write the landing pads onto the terrain
	    for (; x < max; x++){
	      DrawPixel(terrain, rgb.r, rgb.g, rgb.b, x, y);	 
	      DrawPixel(terrain, rgb.r, rgb.g, rgb.b, x, y+1);	 
	      DrawPixel(terrain, rgb.r, rgb.g, rgb.b, x, y+2);	 
	      draw_line(terrain, 80, 80, 80, x, y+3);
	    }

	    x--;
	    level.num_landings++;
	  }
	
	  yd = -1;
	  miny = ( (random()%(TERRAIN_YSIZE/2)) + 10);
	  maxy = ( (random()%(TERRAIN_YSIZE/2) + TERRAIN_YSIZE/2) - 10 );

	}

      }

      if (yd == -1){
	if (y < miny){
	  yd = 1;
	  miny = ( (random()%(TERRAIN_YSIZE/2)) + 10);
	  maxy = ( (random()%(TERRAIN_YSIZE/2) + TERRAIN_YSIZE/2) - 10 );
	
	}
      }

      y = y + yd;

    }

    SDL_UnlockSurface(terrain);
    sprite = new_sprite_surface(terrain, 0, TERRAIN_YSIZE, 1);
    level.sprite = sprite;

    level.fuel = (700 - (game.difficulty * 25));

    if (level.fuel < 300){
      level.fuel = 300;
    }

  }

  return(level);
}

void draw_score(Game game, SDL_Surface *screen, Sprite ship){
  char display_string[100];

  sprintf(display_string, "Fuel %d", game.fuel); 
  
  if (game.small_font > -1){
    DT_DrawText(display_string, screen, game.small_font, 1, 1 );
  }

  sprintf(display_string, "X Velocity %.2f", ship.x_vel); 

  if (game.small_font > -1){
    DT_DrawText(display_string, screen, game.small_font, 100, 1);
  }

  sprintf(display_string, "Y Velocity %.2f", ship.y_vel); 

  if (game.small_font > -1){
    DT_DrawText(display_string, screen, game.small_font, 250, 1);
  }

  sprintf(display_string, "Score: %d", game.score); 

  if (game.small_font > -1){
    DT_DrawText(display_string, screen, game.small_font, 500, 1);
  }


}


void game_over(Game *game, Sprite background, Sprite ship, SDL_Surface *screen, int first_time){

  SDL_Event event;

  draw_sprite(screen, background);

  /* if they just lost, show them thier score and stuff */

  if (first_time == 0){
    DT_DrawText("Game Over", screen, game->big_font, 215, 50 ); 

    draw_score(*game, screen, ship);


  }

  /* reset all the stuff for a new game */

  game->score = 0;
  game->difficulty = 0;
  game->ships_remaining = 3;
  game->gravity = 0.05;

  /* show the game name, etc */

  DT_DrawText("Moon Lander", screen, game->big_font, 200, 25 );
  DT_DrawText("Arrow keys control the ship, q quits", screen, game->big_font, 25, 75 );
  DT_DrawText("press any other key to continue", screen, game->big_font, 75, 100 );

  SDL_Flip(screen);

  /* pause for half a sec, then flush the buffer.
     this is in case they were still holding down the keys
     when they crashed */

  SDL_Delay(500);
  
  while (SDL_PollEvent(&event)){};

  press_any_key(SDLK_q);
}


Level new_round(Sprite *ship, SDL_Surface *screen, Game *game,
		int died_won, Level current_level, int count,
		Sprite miniship, Sprite background) {

  SDL_Event event;
  Level level;
  char display_string[100];

  /* clear event buffer */

  while (SDL_PollEvent(&event)){};

  /* reset ship */
  ship->x = XSIZE / 2;
  ship->y = 1;
  ship->y_vel=0;
  ship->x_vel=0;

  
  if (died_won == 0){  /* died */
    game->ships_remaining--;
    game->fuel = current_level.fuel;
    level = current_level;
  }
  else if (died_won == 1) {  /* won */
    game->score += current_level.landing_score[count];
    game->difficulty++;
    
    /* get new level */

    level = random_level(*game);

    game->fuel = level.fuel;
    game->gravity = 0.05 + (game->difficulty * 0.001 );
  }
  else if (died_won == 3) { /* starting new game */
    level = random_level(*game) ;
    game->fuel = level.fuel;
  }



  /* draw start round screen */

  draw_sprite(screen, background);
  remaining_ships( screen, miniship, *game );
  draw_sprite(screen, *ship);
  draw_sprite(screen, level.sprite);

  /* message 1  */

  draw_score(*game, screen, *ship);

  if (game->difficulty == 4) {
    sprintf(display_string, "Level: %d", (game->difficulty) + 1);
    DT_DrawText("Not Bad", screen, game->big_font, 250, 250 );
    DT_DrawText(display_string, screen, game->big_font, 250, 150 );
  }

  else if (game->difficulty == 9){
    sprintf(display_string, "Level: %d", (game->difficulty) + 1);
    DT_DrawText("You think you're hot shit, huh?", screen, game->big_font, 125, 250 );
    DT_DrawText(display_string, screen, game->big_font, 250, 150 );
  }

  else if (game->difficulty == 14){
    sprintf(display_string, "Level: %d", (game->difficulty) + 1);
    DT_DrawText("Starfleet called, they want to", screen, game->big_font, 100, 235 );
    DT_DrawText("offer you a job", screen, game->big_font, 174, 260 );
    DT_DrawText(display_string, screen, game->big_font, 250, 150 );
  }

  else if (game->difficulty == 19){
    sprintf(display_string, "Level: %d", (game->difficulty) + 1);
    DT_DrawText("The force is strong in you", screen, game->big_font, 125, 250 );
    DT_DrawText(display_string, screen, game->big_font, 250, 150 );
  }
  
  else {
    sprintf(display_string, "Level: %d", (game->difficulty) + 1);
    DT_DrawText(display_string, screen, game->big_font, 250, 150 );
  }

  /* flip screen so they show up */

  SDL_Flip(screen);

  SDL_Delay(2500);

  /* message 2  */

  draw_sprite(screen, background);
  remaining_ships( screen, miniship, *game );
  draw_sprite(screen, *ship);
  draw_sprite(screen, level.sprite);

  draw_score(*game, screen, *ship);

  DT_DrawText("Ready...", screen, game->big_font, 250, 150 );



      /* flip screen so they show up */

  SDL_Flip(screen);

  SDL_Delay(1800);

  draw_sprite(screen, background);
  remaining_ships( screen, miniship, *game );
  draw_sprite(screen, *ship);
  draw_sprite(screen, level.sprite);
  // Draw in the remaining ships

  /* display fuel  */

  draw_score(*game, screen, *ship);

  DT_DrawText("GO!", screen, game->big_font, 290, 150 );

      /* flip screen so they show up */

  SDL_Flip(screen);

  SDL_Delay(500);

  return(level);

  
}


void explode (SDL_Surface *screen, 
	      Sprite exploder, 
	      Sprite *explosion, 
	      Sprite terrain, 
	      Sprite background, 
	      int ActualTime, 
	      int LastTime,
	      Sprite miniship,
	      Game game
	      ) {

  int count;

#ifndef NOSOUND
  Mix_Chunk *wave;
  Mix_HaltChannel(1);
  wave = new_audio("sounds/explosion2.wav");
  play_audio(wave, 0);
#endif

  for (count = 1; count < 29; count++){
    explosion[count-1].x = ( (exploder.x + (exploder.w / 2)) - (explosion[count-1].w / 2) );
    explosion[count-1].y = ( (exploder.y + (exploder.h / 2)) - (explosion[count-1].h / 2) );

    draw_sprite(screen, background);
    remaining_ships( screen, miniship, game );
    draw_sprite(screen, terrain);

    if (count < 28) {
      draw_sprite(screen, explosion[count-1]);
    }
    
    /* display fuel  */

    draw_score(game, screen, exploder);

    // Draw in the remaining ships

    /* Update the screen */
    
    SDL_Flip(screen);

    /* timer - so that things run at an even speed regardless of cpu speed */

    ActualTime = SDL_GetTicks();
    if (ActualTime < LastTime + FPS){
        SDL_Delay(LastTime + FPS - ActualTime);
    }
    LastTime = ActualTime;

 
  }

}

void win (SDL_Surface *screen, Game game) {

#ifndef NOSOUND
  Mix_Chunk *wave;
  Mix_HaltChannel(1);
  wave = new_audio("sounds/victory.wav");
  Mix_VolumeChunk(wave, 80);
  play_audio(wave, 0); 
#endif

  if (game.big_font > -1){
    DT_DrawText("Excellent", screen, game.big_font, (XSIZE / 2) - 66, (YSIZE / 3) - 9 );
  }
  
  /* Update the screen */

  SDL_Flip(screen);
  SDL_Delay(1000);
}

int main(int argc, char **argv) {
  
  SDL_Surface *screen;
  Sprite ship;
  Sprite background;
  Sprite thrust;
  Sprite thrustb;
  Sprite thrust_left;
  Sprite thrust_right;
  Sprite miniship;
  Sprite explosion[27];

  int ActualTime, LastTime;
  short int collision;
  int down;
  int left;
  int right;
  int count;
  int engine_on=0;
  int engine_on_past=0;
  char name[100];
  char display_string[100];
  int restart=0;
  int won=0;
  int odd_even=0;
  Uint8 *key_table;
  SDL_Event event;
  Level current_level;
  int pause = 0;
  int pressed = 0;
  int x;

  Game game;

#ifndef NOSOUND
  Mix_Chunk *engine;
  printf( "Sound active\n" );
#endif

#ifdef NOSOUND
  printf( "Sound disabled\n" );
#endif


  game.score = 0;
  game.difficulty = 0;
  game.ships_remaining = 3;
  game.gravity = 0.05;

  /* initialize SDL */

  screen = init_sdl(XSIZE, YSIZE, "Moon Lander");

  printf("screen initialized\n");

  /* initialize sprites */

  miniship = new_sprite("images/miniship2.bmp", 0, 0, 1 );
  background = new_sprite("images/back02.bmp", 0, 0, 0);
  ship = new_sprite("images/ship2.bmp", XSIZE / 2, 1, 1);
  thrust = new_sprite("images/thrust_down2.bmp", 0, 0, 1);
  thrustb = new_sprite("images/thrust_down2b.bmp", 0, 0, 1);
  thrust_left = new_sprite("images/thrust_left.bmp", 0, 0, 1);
  thrust_right = new_sprite("images/thrust_right.bmp", 0, 0, 1);


#ifndef NOSOUND
  engine = new_audio("sounds/jet_lp.wav");
  Mix_VolumeChunk(engine, 40);
#endif


  /* intitalize timer */

  ActualTime = SDL_GetTicks();
  LastTime = ActualTime;

  /* load fonts */

  game.small_font = DT_LoadFont("fonts/ConsoleFont.bmp", TRANS_FONT) ; 
  game.big_font = DT_LoadFont("fonts/LargeFont.bmp", TRANS_FONT) ; 

  current_level = random_level(game);

  /* load explosion sprites here, because it takes a second or two*/

  for (count = 1; count < 28; count++){
    sprintf(name, "images/exp1/explosion%02d.tga", count);
    explosion[count -1] = new_sprite(name, 0,0,1); 
  }

  game_over(&game, background, ship, screen, 1);
  
  

  current_level = new_round(&ship, screen, &game, 3, 
			    current_level, count, miniship, background);
  
  while(1){ /* start new round */

    


    /* ---------------------- gameloop ---------------------------- */

    restart = 0;

    while (restart == 0) {
       /* zero out "show thrust" vars */

      down=0;
      left=0;
      right=0;
      engine_on=0;
      won = 0;
      odd_even++;

      /* gravity */

      if( !pause )
	ship.y_vel = ship.y_vel + game.gravity;

      /* read keyboard */

      SDL_PollEvent(&event);

      if (event.type == SDL_QUIT){
	exit(0);
      }

      key_table = SDL_GetKeyState(NULL);
    
      if (key_table[SDLK_RIGHT]){

	if (game.fuel > 0){
	  ship.x_vel = ship.x_vel + 0.07;
	  right = 1;
	  game.fuel = game.fuel - 1;
	  engine_on=1;
	
	}
      }
	    
      if (key_table[SDLK_LEFT]){

	if (game.fuel > 0){
	  ship.x_vel = ship.x_vel - 0.07;
	  left = 1;
	  game.fuel = game.fuel - 1;
	  engine_on=1;
	}
      }

      if (key_table[SDLK_UP]){

	if (game.fuel > 1){
	  ship.y_vel = ship.y_vel - 0.10;
	  down = 1;
	  game.fuel = game.fuel - 2;
	  engine_on=1;
	}
      }
	    
      if (key_table[SDLK_q]){
	exit(0);
      }

      // Toggle pause on/off
      if( key_table[SDLK_p] ) {
	if( !pressed ) {
	  pause = (pause)?0:1;
	  pressed++;
	  if( pause ) {
	    /* printf( "Pause pressed\n" ); */
	    sprintf(display_string, "**PAUSED**"); 
	    
	    if (game.small_font > -1){
	      DT_DrawText(display_string, screen, game.small_font, 300, 20);
	    }
	  }
	}
      }
      else{ 
	if( pressed ) {
	  /* printf( "Pause released\n" ); */
	  pressed = 0;
	}
      }

      if( !pause ) {
#ifndef NOSOUND
	if ( (engine_on == 1) && (! Mix_Playing(1) ) ){
	  Mix_PlayChannel(1, engine, 0);
	}

	if ( (engine_on == 0) && (engine_on_past == 1) ){
	  Mix_HaltChannel(1);
	}
#endif

	engine_on_past = engine_on;

	/* update ship position */

	ship.x = ship.x + ship.x_vel;
	ship.y = ship.y + ship.y_vel;

	/* update rocket exhaust poition */

	thrust.x = (ship.x + ( (ship.w - thrust.w) / 2));
	thrust.y = (ship.y + (ship.h -3));

	thrustb.x = (ship.x + ( (ship.w - thrust.w) / 2));
	thrustb.y = (ship.y + (ship.h -3));

	thrust_right.x = (ship.x - thrust_right.w) ;
	thrust_right.y = ( (ship.y + ship.h) - (thrust_right.h + 8) );

	thrust_left.x = (ship.x + ship.w) ;
	thrust_left.y = ( (ship.y + ship.h) - (thrust_left.h + 8) );


	/* draw screen */

	draw_sprite(screen, background);
	remaining_ships( screen, miniship, game );
	draw_sprite(screen, ship);
	draw_sprite(screen, current_level.sprite);

	// Draw in the remaining ships

	// Write in the landing pads as red if you're going too fast.
	for (count = 0; count < current_level.num_landings; count++){
	  if( ship.y_vel > current_level.landing_speed[count] ) {

	    for( x = current_level.landing_x[count]; 
		 x < current_level.landing_x[count] + current_level.landing_w[count]; x++){

	      DrawPixel(screen, 255, 0, 0, x, current_level.landing_y[count]);	 
	      DrawPixel(screen, 255, 0, 0, x, current_level.landing_y[count]+1);	 
	      DrawPixel(screen, 255, 0, 0, x, current_level.landing_y[count]+2);	 
	    }
	  }
	}

	if (right == 1){
	  draw_sprite(screen, thrust_right);
	}

	if (down == 1){
	  if ( (odd_even%2) == 0){
	    draw_sprite(screen, thrust);
	  }
	  else{
	    draw_sprite(screen, thrustb);
	  }
	}

	if (left == 1){
	  draw_sprite(screen, thrust_left);
	}    

	/* display fuel  */

	draw_score(game, screen, ship);

	for (count = 0; count < current_level.num_landings; count++){
	  sprintf(display_string, "%d", current_level.landing_score[count] ); 

	  if (game.small_font > -1){
	    DT_DrawText(display_string, screen, game.small_font, ( current_level.landing_x[count] + 5 ), 
			( current_level.landing_y[count] + 5 ));
	  }

	}

	/* flip screen so they show up */

	SDL_Flip(screen);

	/* check if off screen and way low */

	if (ship.y > YSIZE) {
	  explode(screen, ship, explosion, current_level.sprite, 
		  background, ActualTime, LastTime, miniship, game );

	  if( !game.ships_remaining ) {	      
	    game_over(&game, background, ship, screen, 0);
	    current_level = new_round(&ship, screen, &game, 3, 
			    current_level, count, miniship, background);
	    restart=1;

	  }
	  else {
	    current_level = new_round(&ship, screen, &game, 0, 
				      current_level, count, miniship, background);
	  }
	}

	/* collision detection */

	collision = collision_detect_perfect(ship, current_level.sprite);

	if (collision > 0){
  
	  /* landing detection */

	  for (count = 0; count < current_level.num_landings; count++){

	    if ( ((ship.x + 2) > current_level.landing_x[count]) 
		 && ( ((ship.x - 2) + ship.w) < 
		      (current_level.landing_x[count] + current_level.landing_w[count]) )  
		 && (ship.y_vel < current_level.landing_speed[count]) 
		 && (ship.x_vel <0.5) && (ship.x_vel > -0.5)){

	      // Bonus ship every 5000 points
	      if( (game.score / 5000) < 
		  ((game.score + current_level.landing_score[count]) / 5000 ) ) {
		// (MLH) This would be a good place to play a sound
		game.ships_remaining++;
	      }
	      win(screen, game);
	      won = 1;
	      restart=1;
	      current_level = new_round(&ship, screen, &game, 1, 
					current_level, count, miniship, background);
	      break;
	    }
		
	  }


	  if (won == 0){

	    /* didn't land */

	    explode(screen, ship, explosion, current_level.sprite, background, 
		    ActualTime, LastTime, miniship, game );

	    if( !game.ships_remaining ) {
	      game_over(&game, background, ship, screen, 0);
	      current_level = new_round(&ship, screen, &game, 3, 
			    current_level, count, miniship, background);
	      restart=1;
	    }
	    else {
	      current_level = new_round(&ship, screen, &game, 0, 
					current_level, count, miniship, background);
	    }
	
	  }
	}
      }    
      else {
	SDL_Flip(screen);
      }

      /* good place to print out debugging stuff */


      /* timer - so that things run at an even speed regardless of cpu speed */

      ActualTime = SDL_GetTicks();
      if (ActualTime < LastTime + FPS){
	SDL_Delay(LastTime + FPS - ActualTime);
      }
      LastTime = ActualTime; 
    }
  
  }

  return(0);
}
